#include <cmath>

#include "canvas/Tree.hpp"
#include "canvas/Circle.hpp"
#include "Math.hpp"

#include <raylib.h>
#include <raymath.h>

#define ITER_PER_FRAME 5000

// Public
void Tree::init(int size)
{
  this->size = size;
  start.x = size / 2;
  start.y = size;
}

void Tree::draw(Dna *dna)
{
  Circle::setSoftEdge(false);

  m_dna = dna;
  draw_calls.push_back({start, 0, (float)size / 4, 1});
  tick();
}

bool Tree::tick()
{
  size_t i = 0;
  while (!draw_calls.empty())
  {
    drawBranch();
    draw_calls.pop_front();
    i++;
    if (i >= ITER_PER_FRAME)
      break;
  }

  return draw_calls.empty();
}

// Private

Vector2 Tree::drawLine()
{
  DrawArgs arg = draw_calls.front();

  arg.angleDeg += 180.0f;
  float angle = (arg.angleDeg * PI) / 180.0f;
  float nx = arg.lenghth * std::sin(angle);
  float ny = arg.lenghth * std::cos(angle);
  Vector2 end = {arg.start.x + nx, arg.start.y + ny};

  float thick = 2.0;
  float fstep = 1.0 / ((arg.lenghth / thick) * 1.5);

  Color colorParent = {
      m_dna->branches[arg.dep - 1].colorR,
      m_dna->branches[arg.dep - 1].colorG,
      m_dna->branches[arg.dep - 1].colorB,
      255};

  Color colorM = {
      m_dna->branches[arg.dep].colorR,
      m_dna->branches[arg.dep].colorG,
      m_dna->branches[arg.dep].colorB,
      255};

  for (float i = 0; i < 1; i += fstep)
  {
    Vector2 point = Vector2Lerp(arg.start, end, i);
    Color color = ColorLerp(colorParent, colorM, i);
    DrawCircleV(point, thick, color); // Fester on the phone to call DrawCircle insted of the Circle shader
    // Circle::setColor(color);
    // Circle::draw(point.x, point.y, thick); // TODO Change to BeginShaderMode and EndShaderMode only onece
  }
  return end;
}

inline uint8_t get_num_of_branches(uint8_t count)
{
  return ((float)count / 255.0f) * 3 + 1;
}

void Tree::drawBranch()
{
  DrawArgs arg = draw_calls.front();
  if (arg.dep >= MAX_DEPTH)
    return;

  Vector2 next = drawLine();

  float next_len = 0.7f;
  float sectors = get_num_of_branches(m_dna->branches[arg.dep].branch_count) + 1;
  float degres = 180.0f / sectors;

  for (size_t i = 0; i < get_num_of_branches(m_dna->branches[arg.dep].branch_count); i++)
  {
    float newAngle = arg.angleDeg - 90 + (degres * (i + 1));
    draw_calls.push_back({next, newAngle, arg.lenghth * next_len, arg.dep + 1});
  }
}
